# encoding: utf-8

"""Release information about the Turbo-Mail mail queue."""

version = "1.1"
description = "Multi-threaded mail queue manager for TurboGears applications."
author = "Top Floor Computer Systems Ltd."
email = "webmaster@topfloor.ca"
url = "http://www.topfloor.ca/turbomail/"
download_url = "http://www.topfloor.ca/turbomail/"
copyright = "© 2006 Top Floor Computer Systems Ltd."
license="MIT"



# encoding: utf-8

"""This module contains only the primary SMTP Dispatch class."""

import logging
log = logging.getLogger("turbomail.dispatch")

from smtplib import SMTP


__all__ = ['Dispatch']


class Dispatch(object):
	"""SMTP message dispatch class.
	
	An instance of the Dispatch class is created for each SMTP
	connection.  Usually, this means one Dispatch instance per
	running thread.
	
	Example usage::
	
		import turbomail
		dispatch = turbomail.Dispatch("localhost")
		message = turbomail.Message(
				"from@localhost",
				"to@localhost",
				"Subject"
			)
		message.plain = "Hello world!"
		dispatch(message)
	
	"""

	def __init__(self, server, username=None, password=None, tls=None, debug=False):
		"""Initialize the Dispatch class.
		
		Authentication is only performed if both I{username} and
		I{password} are not None.  An instance of Dispatch is callable.
		
		@param server: The server (with optional port number) to
		               connect this instance to.
		@type server: string
		
		@param username: The username to use during authentication.
		                 I{Optional.}
		@type username: string
		
		@param password: The password to use during authentaction.
                         I{Optional.}
		@type password: string
		
		@param debug: Enable SMTP verbose logging.  This outputs all
		              communications between the client and server.
		@type debug: bool
		"""
		
		super(Dispatch, self).__init__()
		
		self.server = server
		self.username = username
		self.password = password
		self.tls = tls
		self.debug = debug
		self.log = None
		
		log.debug("Creating SMTP object.")
		self.connection = SMTP()
		self.connection.set_debuglevel(debug)

	@property
	def connected(self):
		"""Return the current SMTP connection status."""
		
		return getattr(self.connection, 'sock', None) is not None

	def connect(self):
		"""Connect to the SMTP server if not already connected.
		
		This process also automatically enables TLS, if available, and
		authenticates against the username and password previousally
		provided.
		"""
		
		if self.connected: return
		
		log.debug("Connecting to SMTP server %s." % self.server)
		self.connection.connect(self.server)

		if self.tls or self.tls is None:
			self.connection.ehlo()
			
			if self.connection.has_extn('STARTTLS') or self.tls:
				self.connection.starttls()
				self.connection.ehlo()
				log.debug("TLS enabled on SMTP server.")
				self.tls = True
			
			else:
				log.debug("TLS not available on SMTP server.")
				self.tls = False

		if self.username and self.password:
			log.debug("Authenticating as %s." % self.username)
			self.connection.login(self.username, self.password)
	
	def disconnect(self):
		"""Disconnect from the SMTP server if connected."""
		
		if not self.connected: return
		
		log.debug("Closing SMTP connection.")
		self.connection.quit()
	
	def __call__(self, message):
		"""Deliver a message via the current SMTP connection.
		
		Calling an instance of the Dispatch class will automatically
		connect, if needed, and will also automatically disconnect if
		debug mode has been enabled.
		
		@param message: This paramater must be a callable which returns
		                a tuple of (from, to, message), where all three
		                are strings, and message is valid content for an
		                e-mail message.
		@type message: callable
		"""
		
		self.connect()
		
		pack = message()
		
		log.info(
				'%(server)s %(user)s %(tls)s "%(addrfrom)s" "%(addrto)s" %(size)d "%(subject)s"' % dict(
						user=['-', self.username][self.username and self.password],
						server=self.server,
						tls=['-', 'TLS'][self.tls],
						addrfrom=pack[0],
						addrto=pack[1],
						size=len(pack[2]),
						subject=message._message['Subject']
					)
			)
		self.connection.sendmail(*pack)
		
		if self.debug: self.disconnect()



# encoding: utf-8

"""Generic Pool and MailPool class definitions."""

import logging
log = logging.getLogger("turbomail.pool")

import math
from Queue import Queue, Empty
from threading import Event, Thread
from turbomail.dispatch import Dispatch


__all__ = ['Pool', 'MailPool']


class Pool(Thread):
	"""A threadpool which checks regularily for new jobs and spawns processes
	as needed.
	
	Do not use this class directly.  Always subclass and override the worker
	method.
	"""
	
	def __init__(self, interval=10, threads=4, jobs=10, timeout=60, **kw):
		"""Initialize the threadpool.
		
		@param interval: A delay, in seconds, between spawn runs.
		@type interval: int
		
		@param threads: The maximum number of concurrent threads.
		@type threads: int
		
		@param jobs: The maximum number of jobs a single thread is
		             allowed to handle before dying of old age.
		@type jobs: int
		
		@param timeout: The amount of time, in seconds, a thread is
		                allowed to sit idle before dying of starvation.
		@type timeout: int
		"""
		
		super(Pool, self).__init__()
		
		self._pool = 0
		self._queue = Queue()
		self._finished = Event()
		self._interval = interval
		self._threads = threads
		self._jobs = jobs
		self._timeout = timeout
		self._kw = kw
		
		log.debug("Thread pool created.")
	
	def enqueue(self, work):
		"""Enqueue a Message instance.
		
		@todo: Optionally, this method should immediately trigger the
		       management thread, allowing for almost immediate latency.
		
		@param work: The unit of work can be any callable that returns a
		             three-item tuple containing the sender and
		             recipient addresses and a properly formatted MIME
		             message, in that order.  The preferred type is an
		             instance of the Message class or subclass.
		@type work: callable
		"""
		
		log.debug("Work enqueued.")
		self._queue.put(work)
	
	def shutdown(self):
		"""Quit the management thread and shutdown the queue."""
		
		log.debug("Shutdown requested.")
		self._finished.set()

	def spawn(self):
		thread = Thread(target=self.wrapper)
		thread.start()
		self._pool += 1

	def run(self):
		"""The management thread.
		
		Do not call directly.  Instead, use the I{start} method.
		"""
		
		log.debug("Beginning thread pool main loop.")
		
		while True:
			if self._finished.isSet():
				log.debug("Shutdown request acknowledged.")
				break
			
			if not self._queue.empty():
				log.debug("Estimate %d work units in the queue." % self._queue.qsize())
				
			optimum_threads = min(self._threads, math.ceil(self._queue.qsize() / float(self._jobs)))
			
			if not self._queue.empty() and self._pool < optimum_threads:
				log.debug("Creating %d threads." % (optimum_threads - self._pool))
				for i in xrange(int(optimum_threads - self._pool)):
					self.spawn()
					
			self._finished.wait(self._interval)
			
		log.debug("Thread pool main loop has ended.")
	
	def wrapper(self):
		"""Thread wrapper to log and keep track of the active thread count."""
		
		log.debug("Thread pool worker starting up.")
		
		self.worker()
		
		self._pool -= 1
		log.debug("Thread pool worker finished.")

	def worker(self):
		"""This method must be overridden in a subclass and is used to
		perform the work of the threadpool.
		
		Will raise a NotImplementedError exception if not subclassed."""
		
		raise NotImplementedError


class MailPool(Pool):
	"""Mail delivery threadpool.
	
	This class delivers messages from a queue using the Dispatch class.

	Example usage::
	
		import turbomail
		pool = turbomail.MailPool()
		message = turbomail.Message(
				"from@localhost",
				"to@localhost",
				"Subject"
			)
		message.plain = "Hello world!"
		pool.enqueue(message)

	"""
	
	def worker(self):
		"""Deliver up to I{jobs} messages per queue.
		
		If there are no messages available in the queue, the worker
		will wait up to I{timeout} seconds for data.  If the timeout
		expires, the thread will exit gracefully."""
		
		count = 0
		dispatch = Dispatch(**self._kw)
		
		log.debug("Worker starting work.")
		
		while True:
			if not count < self._jobs:
				log.debug("Worker death from old age - spawning child.")
				self.spawn()
				break
			
			try:
				unit = self._queue.get(True, self._timeout)
				dispatch(unit)

			except Empty:
				log.debug("Worker death from starvation.")
				break
			
			count += 1



# encoding: utf-8

"""Exceptions used by TurboMail to report common errors."""


__all__ = ['MailException', 'MailNotEnabledException', 'MailConfigurationException']


class MailException(Exception):
	"""The base for all TurboMail exceptions."""
	
	pass

		
class MailNotEnabledException(MailException):
	"""Attempted to use TurboMail before being enabled."""
	
	def __str__(self):
		return "An attempt was made to use a facility of the TurboMail " \
			   "framework but outbound mail hasn't been enabled in the" \
			   "config file [via mail.on]."
	
	
class MailConfigurationException(MailException):
	"""There was an error in the configuration of TurboMail."""
	
	args = ()
	def __init__(self, message):
		self.message= message
		
	def __str__(self):
		return self.message



# encoding: utf-8

"""TurboGears extension startup and shutdown interface."""

import logging
log = logging.getLogger("turbomail")

import turbogears

import turbomail
from turbomail.exceptions import *
from turbomail.pool import MailPool


__all__ = ['start_extension', 'shutdown_extension']


def start_extension():
	"""TurboGears extension startup.
	
	Exits immediately if TurboMail is not enabled and creates a MailPool
	instance if all is well.
	"""
	
	if not turbogears.config.get("mail.on", False):
		return
	
	if turbogears.config.get("mail.server", None) is None:
		raise MailConfigurationException("Outbound server must be specified.")
	
	log.info("Outbound mail queue manager starting.")
	
	turbomail.queue = MailPool(
			interval=turbogears.config.get("mail.interval", 10),
			threads=turbogears.config.get("mail.threads", 4),
			jobs=turbogears.config.get("mail.jobs", 10),
			timeout=turbogears.config.get("mail.timeout", 60),
			server=turbogears.config.get("mail.server"),
			username=turbogears.config.get("mail.username", None),
			password=turbogears.config.get("mail.password", None),
			debug=turbogears.config.get("mail.debug", False),
			tls=turbogears.config.get("mail.tls", None)
		)
	turbomail.queue.start()
	turbomail.enqueue = turbomail.queue.enqueue


def shutdown_extension():
	"""TurboGears extension shutdown.
	
	Exits immediately if TurboMail is not enabled and shuts down the
	MailPool object safely.
	"""
	
	if not turbogears.config.get("mail.on", False):
		return

	log.info("Outbound mail queue manager shutting down.")
	
	turbomail.enqueue = None
	turbomail.queue.shutdown()
	turbomail.queue = None



# encoding: utf-8

"""MIME-encoded electronic mail message classes."""

import turbogears, re

from email.Utils import formatdate
from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Encoders import encode_base64


__all__ = ['Message', 'TemplateMessage']

_rich_to_plain = re.compile(r"(<[^>]+>)")


class Message(object):
	"""Simple e-mail message class.
	
	Message provides a means to easily create e-mail messages to be
	sent through the Dispatch mechanism or MailPool.  Message provides
	various helper functions to correctly format plain text, dual plain
	text and rich text MIME encoded messages, as well as handle
	embedded and external attachments.
	
	Example usage::
	
		import turbomail
		message = turbomail.Message(
				"from@host.com",
				"to@host.com",
				"Subject"
			)
		message.plain = "This is a plain message."
	
	@ivar plain: The plain text content of the message.
	@type plain: string
	
	@ivar rich: The rich text (HTML) content of the message.  Plain text
	            content B{must} be available as well.
	@type rich: string
	"""
	
	
	def __init__(self, sender, recipient, subject, encoding=None):
		"""Instantiate a new Message object.
		
		An instance of Message is callable.
		
		@param sender: The e-mail address of the sender.  This is
					   encoded as the "From:" SMTP header.
		@type sender: string
		
		@param recipient: The recipient of the message.	 This gets
						  encoded as the "To:" SMTP header.
		@type recipient: string
		
		@param subject: The subject of the message.	 This gets encoded
						as the "Subject:" SMTP header.
		@type subject: string
		
		@param encoding: The encoding to send the text parts of the
		                 message contents as.  Note that this does not
		                 perform any conversion for you, it simply
		                 passes this value to MIMEText().
		@type encoding: string
		"""
		
		super(Message, self).__init__()
		
		self._message = MIMEMultipart() # ('related')
		self._processed = False
		self._attachments = []
		self._charset = [turbogears.config.get("mail.encoding", 'us-ascii'), encoding][encoding is None]
		
		self.plain = None
		self.rich = None

		self._message['From'] = sender
		self._message['To'] = recipient
		self._message['Date'] = formatdate(localtime=True)
		self._message['Subject'] = subject
			
	def attach(self, file, name=None):
		"""Attach an on-disk file to this message.
		
		@param file: The path to the file you wish to attach.
		@type file: string
		
		@param name: You can optionally override the filename of the
		             attached file.  This name will appear in the
		             recipient's mail viewer.  B{Optional.}
		@type name: string
		"""
		
		part = MIMEBase('application', "octet-stream")

		fp = open(file, "rb")
		part.set_payload(fp.read())
		Encoders.encode_base64(part)

		if name:
			part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(name))
		else:
			part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(file))
			
		self._attachments.append(part)
	
	def embed(self, file, name):
		"""Attach an on-disk image file and prepare for HTML embedding.
		
		This method should only be used to embed images.
		
		@param file: The path to the file you wish to attach.
		@type file: string
		
		@param name: The CID name to use for the embedded object.
		@type name: string
		"""
		
		from email.MIMEImage import MIMEImage
		
		fp = open(file, 'rb')
		part = MIMEImage(fp.read())
		fp.close()
		
		part.add_header('Content-ID', '<%s>' % name)
		
		self._attachments.append(part)

	def _process(self):
		"""Produce the final MIME message.
		
		Additinoally, if only a rich text part exits, strip the HTML to
		produce the plain text part.  (This produces identical output as
		KID.)
		"""
		
		if self.rich and not self.plain:
			self.plain = _rich_to_plain.sub('', self.rich)
		
		if self.plain and self.rich:
			alternative = MIMEMultipart('alternative')
			self._message.attach(alternative)
			
			alternative.attach(MIMEText(self.plain, _charset=self._charset))
			alternative.attach(MIMEText(self.rich, 'html', _charset=self._charset))
		
		else:
			self._message.attach(MIMEText(self.plain, _charset=self._charset))
		
		for attachment in self._attachments:
			self._message.attach(attachment)
		
		self._processed = True 
	
	def __call__(self):
		"""Produce a valid MIME-encoded message and return valid input
		for the Dispatch class to process.

		@return: Returns a tuple containing sender and recipient e-mail
		         addresses and the string output of MIMEMultipart.
		@rtype: tuple
		"""
		
		if not self._processed:
			self._process()
		
		return (self._message['From'], self._message['To'], self._message.as_string())


class TemplateMessage(Message):
	"""A template which accepts a named template with arguments.
	
	Example usage::
	
		import turbomail
		message = turbomail.TemplateMessage(
				"from@host.com",
				"to@host.com",
				"Subject",
				"app.templates.mail",
				dict()
			)
	
	Do not specify message.plain or message.rich content - the template
	will override what you set.  If you wish to hand-produce content,
	use the Message class.
	"""
	
	def __init__(self, sender, recipient, subject, template, variables={}, encoding=None):
		"""Store the additonal template and variable information.
		
		@param template: A dot-path to a valid KID template.
		@type template: string
		
		@param variables: A dictionary containing named variables to
		                  pass to the template engine.
		@type variables: dict
		"""
		
		self._template = template
		self._variables = dict(sender=sender, recipient=recipient, subject=subject)
		self._variables.update(variables)
				
		super(TemplateMessage, self).__init__(sender, recipient, subject, encoding)
	
	def _process(self):
		"""Automatically generate the plain and rich text content."""
		
		turbogears.view.base._load_engines()
		
		localvars = dict()
		
		for (i, j) in self._variables.iteritems():
			if callable(j):
				localvars[i] = j()
			else:
				localvars[i] = j
		
		render = turbogears.view.engines.get('kid').transform(localvars, self._template)
		render = turbogears.widgets.base.PlainHTML(elements=render)
		
		self.plain = render.serialize(output="plain")			
		self.rich = render.serialize(output="html")
		
		return super(TemplateMessage, self)._process()



# encoding: utf-8

from setuptools import setup, find_packages
from turbogears.finddata import find_package_data

import os
execfile(os.path.join("turbomail", "release.py"))

setup(
		name="TurboMail",
		version=version,
	
		description=description,
		author=author,
		author_email=email,
		url=url,
		download_url=download_url,
		license=license,
	
		install_requires = ["TurboGears >= 1.0b1",],
		zip_safe=True,
		packages=find_packages(),
		package_data = find_package_data(where='turbomail', package='turbomail'),
		keywords = ["turbogears.extension"],
		classifiers = [
			'Development Status :: 5 - Production/Stable',
			'Operating System :: OS Independent',
			'Programming Language :: Python',
			'Topic :: Software Development :: Libraries :: Python Modules',
			'Framework :: TurboGears',
		],
		test_suite = 'nose.collector',
		entry_points = {'turbogears.extensions': ["turbomail = turbomail"]}
	)
	



# encoding: utf-8

"""
Introduction
============

	TurboMail is a TurboGears extension - meaning that it starts up and
	shuts down alongside TurboGears applications you write in the same
	way that visit tracking and identity do.  TurboMail uses built-in
	Python modules for SMTP communication and MIME e-mail creation, but
	greatly simplifies these tasks by performing the grunt-work for you.
	
	Being multi-threaded, TurboMail allows you to enqueue messages to be
	sent and then immediately continue with processing, resulting in a
	much more fluid user experience.  Threads are handled intelligently
	(increasing the number of threads as demand increases) and they are
	automatically recycled.  There is only ever one SMTP connection per
	thread.
	
	Benchmarking
	------------
	
		Throughput using the default options is sufficient for most use:
		100 messages in 45 seconds; just over 2 messages a second. Using
		a greater number of threads, 10 vs. 4, 100 messags take 30
		seconds; just over 3 messages a second.  YMMV.  Note that if a
		thread is idle, it will immediately deliver requests added to
		the queue.
	

Installation
============

	Simply easy_install the package::

		easy_install TurboMail

	TurboMail installs no external scripts.

Upgrade
=======

	Upgrading also uses easy_install::
	
		easy_install -U TurboMail

Configuration
=============

	TurboMail understands a large number of configuration options, all
	piggy-backed from your application's configuration.  Organized into
	two groups, the advanced set can be safely ignored in most
	applications.  Each option is listed with its default value.
	
	Simple Options
	--------------
	
		 - I{mail.on} (Default: B{False}) Enable TurboMail.  B{Required.}
		 - I{mail.server} (Default: B{None}) SMTP server address.
		   B{Required.}
		 - I{mail.username} (Default: B{None})
		 - I{mail.password} (Default: B{None})
	
		Both a username and password are required to enable
		authentication.
	
	Advanced Options
	----------------
	
		 - I{mail.debug} (Default: B{False}) Output all SMTP server
		   communications.
		 - I{mail.interval} (Default: B{10}) Polling delay between new
		   thread creation, in seconds.
		 - I{mail.threads} (Default: B{4}) Maximum number of concurrent
		   threads.
		 - I{mail.jobs} (Default: B{10}) Maximum number of job units per
		   thread.
		 - I{mail.timeout} (Default: B{60}) Maximum time a worker thread
		   will wait for additional jobs, in seconds.
		 - I{mail.tls} (Defalut: None) Enable or disable TLS, None will
		   attempt to auto-detect TLS.  This will not always work.
		 - I{mail.encoding} (Defalut: 'us-ascii') Set the encoding on
		   the MIMEText parts of the message.  This will not perform
		   encoding conversion for you.
		
		In debug mode using a single thread with a maximum of one job
		can be advantageous.  Having a single thread with a maximum of a
		single job limits TurboMail to a single SMTP connection at a
		time and automatically disconnects for I{each message}.

Basic Usage
===========

	To use TurboMail in your TurboGears application, after adding the
	appropriate configuration options to your application, perform the
	following steps:
	
	 1. Import TurboMail::
	
	    	import turbomail
	
	 2. Create a L{Message} or L{TemplateMessage} object::
	
	    	message = turbomail.Message(from, to, subject)
	
	 3. Set some content for your message::
	
	    	message.plain = "Hello world!"
	
	 4. Enqueue your message::
	
	    	turbomail.enqueue(message)
	
	Your message will now have been enqueued.  It will take at most
	I{mail.interval} seconds for a thread to be created to deliver it.
	The best case scenario is if there is an idle thread waiting, in
	which case delivery will be immediate.

Advanced Usage - Logging
========================

	Additionally, you can configure your application to log TurboMail
	events differently, in a more loggable and machine parsable way.
	You do so by adding the following lines to the formatters section
	of your log.cfg::
	
		[[[timed_message]]]
		format='*(asctime)s *(message)s'
	
	Add the following to the handlers section of log.cfg::
	
		[[[mail_out]]]
		class='StreamHandler'
		level='INFO'
		args='(sys.stdout,)'
		formatter='timed_message'
	
	And finally, add the following to your dev.cfg::
	
		[[[mail]]]
		level='INFO'
		qualname='turbomail.dispatch'
		handlers=['mail_out']
		propagate=0
	
	If you wish to log mail dispatch to a file, for example in your
	production configuration, use this instead of the above::
	
		[[handlers]]
		[[[mail_out]]]
		args="('mail.log',)"
		class='StreamHandler'
		level='INFO'
		formatter='timed_message'
		
		[[loggers]]
		[[[mail]]]
		level='INFO'
		qualname='turbomail.dispatch'
		handlers=['mail_out']
		propagate=0	

Changelog
=========

	Version 1.0
	-----------
		 - Initial release.
	
	Version 1.0.1
	-------------
		 - Minor updates to remove unneeded arguments.
		 - Complete source-level epydoc documentation.
	
	Version 1.0.4.1
	---------------
		 - Better auto-detection of TLS capability.
		 - A new configuration directive, mail.tls; True, False, or
		   None to auto-detect.
		 - Fixes a bug in TemplateMessage which rendered it
		   non-functional.
		 - Changed the behavior of a worker dying from old age to spawn
		   a new process immediately.
		 - Minor fixes and updates to the documentation.
		 - Benchmark results in the documentation.
	
	Version 1.0.4.2
	---------------
		 - Added encoding configuration directive.
		 - Encoding can be passed to a Message constructor to override
		   the encoding on a message-by-message basis.
	
	Version 1.1
	-----------
		 - Cleaned up the log output.
		 - Added documentation for logging.
		 - KID i18n session bug fixed in TurboGears 1.0 and trunk SVN.
		 - Marked as stable for the Python Cheese Shop.
		

@var queue: After TurboGears startup within an application which has
            enabled TurboMail, I{queue} is an instance of a MailPool
            object.

@var enqueue: After TurboGears startup within an application which has
              enabled TurboMail, I{enqueue} becomes a bound reference to
              the L{enqueue} method of the I{queue} object.
		
"""

import logging
log = logging.getLogger("turbomail")

from turbomail.startup import start_extension, shutdown_extension
from turbomail.exceptions import *
from turbomail.message import Message, TemplateMessage


__all__ = ['MailException', 'MailNotEnabledException', 'MailConfigurationException', 'Message', 'TemplateMessage', 'enqueue', 'dispatch', 'exceptions', 'message', 'pool', 'startup']


queue = None

def enqueue():
	"""Enqueue a message in the message thread-pool queue.
	
	After TurboGears startup within an application which has enabled
	TurboMail, I{enqueue} becomes a bound reference to the L{enqueue}
	method of the I{queue} object."""
	
	raise MailNotEnabledException
